searchData={"items":[{"type":"behaviour","title":"Lamina","doc":"Lamina allows you to define a run-time configuration pipeline that can merge\nconfiguration from several sources.  This allows the system to be reactive to\nchanges in its environment.","ref":"Lamina.html"},{"type":"behaviour","title":"Example - Lamina","doc":"The following example defines a configuration for an imaginary HTTP server\napplication which takes it's configuration from a combination of default\nvalues, the OTP application environment and system environment variables:\n\n```elixir\ndefmodule MyHttpServer.Config do\n  use Lamina\n\n  provider(Lamina.Provider.Default, listen_port: 4000, listen_address: \"0.0.0.0\")\n  provider(Lamina.Provider.ApplicationEnv, otp_app: :my_http_server, key: MyHttpServer.Endpoint)\n  provider(Lamina.Provider.Env, prefix: \"HTTP\")\n\n  config :listen_port do\n    cast(&Lamina.Cast.to_integer/1)\n\n    validate(fn\n      port when is_integer(port) and (port in [80, 443] or port >= 1000) -> true\n      _ -> false\n    end)\n  end\n\n  config :listen_address do\n    validate(fn\n      address when is_binary(address) ->\n        address\n        |> String.to_charlist()\n        |> :inet.parse_address()\n        |> case do\n          {:ok, _} -> true\n          _ -> false\n        end\n\n      _ ->\n        false\n    end)\n  end\nend\n```\n\nProvider order is preserved, such that providers added later (via the\n`provider/1` or `provider/2` macro) have more priority than their\npredecessors.  This has the effect that when more than one provider can\nprovide a value for a given configuration item, the most preferred value will\nbe returned.\n\nEach configuration item is defined using the `config/1` or `config/2` macro.\nIf the configuration item does not need casting to another type, nor\nvalidation then just defining it with `config/1` is sufficient.  In some cases\nit is necessary to provide additional casting or validating functions.  They\ncan be provided by passing a block containing the `cast/1` or `validate/1`\nmacros.\n\nMake sure that you add your configuration module to your application's\nsupervisor tree **before** any processes that rely on it's information.\nLamina will fail to start or shutdown on any errors it encounters.","ref":"Lamina.html#module-example"},{"type":"behaviour","title":"Lifetimes - Lamina","doc":"All configuration items in Lamina are explicitly marked with a lifetime, which\nmust be specified by the configuration provider when returning values.  The\nsemantics are as follows:\n\n  - `:volatile` - a configuration that could potentially be different every\n    time it is read.  Volatile configuration items are returned by the\n    `ApplicationEnv` and `Env` providers.\n  - `:static` - a configuration value that is not going to change until the\n    provider changes it.  Static configuration items are returned by the\n    `Default` provider, but could also be used for a configuration provider\n    that notifies the system of configuration changes in some way.\n  - `{non_neg_integer(), System.time_unit()}` - a value that has a specific\n    expiry time.  This may be used for a configuration source that has\n    explicit leases on values (ala [Vault](https://www.vaultproject.io/) or a\n    value for which querying is expensive, and providing an expiry would\n    effectively cache it.","ref":"Lamina.html#module-lifetimes"},{"type":"behaviour","title":"Querying - Lamina","doc":"When asked to retrieve a configuration value, Lamina queries it's ETS table\nusing the following query plan; values for which there is no expiry, or which\nhave not yet expired, ordered by provider weight, descending.  It only ever\nreturns a single row.\n\nIf the returned row is marked as `:volatile` then the configuration provider\nis immediately queried for a new value, meaning that these requests will pay\nthe cost of a `GenServer.call/3` to ensure freshness.  If this is an issue\nthen you should consider changing the provider lifetime to use an expiry.  The\n`ApplicationEnv` and `Env` providers have a configuration option to do this.\nIf you are the developer of a volatile provider, it is strongly suggested that\nyou provide for this use case.","ref":"Lamina.html#module-querying"},{"type":"behaviour","title":"Server configuration - Lamina","doc":"The following options can be passed to the `use Lamina` macro, although it's\nprobably advisable to leave them as their defaults.\n\n  - `gc_timeout: pos_integer()` - how long the server should be idle before\n    removing expired configuration from the ETS table in milliseconds.\n    Defaults to 3000.\n  - `ttl_refresh_fraction: float` - when presented with a configuration value\n    which has an expiry, the server queues a refresh at some point prior to\n    the value expiring, in order to avoid having missing configuration.\n    Setting this to a value between `0` and `1` specifies the proportion of\n    the expiry time to wait before attempting to refresh the value.  Defaults\n    to `0.95`.","ref":"Lamina.html#module-server-configuration"},{"type":"behaviour","title":"Configuration subscriptions - Lamina","doc":"Lamina defines a `subscribe/1` and `unsubscribe/1` function on each\nconfiguration module, which uses a `Registry` to handle pub-sub for\nconfiguration changes.\n\nThis allows your processes to subscribe to configuration changes and update or\nrestart any services they provide.\n\n#","ref":"Lamina.html#module-configuration-subscriptions"},{"type":"behaviour","title":"Example - Lamina","doc":"For example, a simple HTTP server which changes it's listen port in response\nto a configuration change:\n\n```elixir\ndefmodule MyHttpServer.Cowboy do\n  use GenServer\n  alias Plug.Cowboy\n  alias MyHttpServer.{Config, Plug}\n\n  def init(_) do\n    with {:ok, port} <- Config.listen_port(),\n        {:ok, srv} <- Cowboy.http(Plug, [], port: port),\n        :ok <- Config.subscribe(:listen_port) do\n      {:ok, %{srv: srv, port: port}}\n    end\n  end\n\n  def handle_info({:config_change, Config, :listen_port, _old_port, new_port}, %{\n        port: current_port\n      })\n      when new_port != current_port do\n    with :ok <- Cowboy.shutdown(Plug.HTTP),\n        {:ok, srv} <- Cowboy.http(Plug, [], port: new_port) do\n      {:noreply, %{port: new_port, srv: srv}}\n    else\n      {:error, reason} -> {:stop, reason, nil}\n    end\n  end\n\n  def handle_info({:config_change, _, _, _}, state), do: {:noreply, state}\nend\n```","ref":"Lamina.html#module-example"},{"type":"callback","title":"Lamina.__lamina__/1","doc":"","ref":"Lamina.html#c:__lamina__/1"},{"type":"callback","title":"Lamina.config_change/3","doc":"Called when the `Lamina.Server` detects a configuration change.\n\nThis is a simple mechanism to allow the server to notify your system that a\nconfiguration value has changed.\n\nSee also `subscribe/1`.","ref":"Lamina.html#c:config_change/3"},{"type":"type","title":"Lamina.config_key/0","doc":"","ref":"Lamina.html#t:config_key/0"},{"type":"module","title":"Lamina.Cast","doc":"A helpful library of casts.\n\nDefining a cast function is brain-dead simple, however sometimes you just\ndon't want to write the same thing over and over.  Here is a helpful pile of\nfunctions for common configuration castings.\n\nFeel free to [open a PR][1] to add more.\n\n[1]: https://harton.dev/james/lamina/-/merge_requests","ref":"Lamina.Cast.html"},{"type":"function","title":"Lamina.Cast.to_atom/1","doc":"Attempt to convert the inbound value into an atom.","ref":"Lamina.Cast.html#to_atom/1"},{"type":"function","title":"Lamina.Cast.to_boolean/1","doc":"Attempt to convert the inbound value into a boolean.\n\nThings that are considered true:\n\n  * A literal `true`.\n  * The words \"true\" or \"yes\" in any capitalisation.\n\nEverything else is false.","ref":"Lamina.Cast.html#to_boolean/1"},{"type":"function","title":"Lamina.Cast.to_float/1","doc":"Attempts to convert the inbound value into an float.","ref":"Lamina.Cast.html#to_float/1"},{"type":"function","title":"Lamina.Cast.to_integer/1","doc":"Attempts to convert the inbound value into an integer.","ref":"Lamina.Cast.html#to_integer/1"},{"type":"function","title":"Lamina.Cast.to_string/1","doc":"Attempt to convert the inbound value into an atom.","ref":"Lamina.Cast.html#to_string/1"},{"type":"module","title":"Lamina.DSL","doc":"Defines the macros used for building a configuration module.","ref":"Lamina.DSL.html"},{"type":"macro","title":"Lamina.DSL.config/1","doc":"Defines an individual configration parameter.\n\nThe same as `config/2`, except that no block is provided.","ref":"Lamina.DSL.html#config/1"},{"type":"macro","title":"Lamina.DSL.config/2","doc":"Defines an individual configration parameter.","ref":"Lamina.DSL.html#config/2"},{"type":"macro","title":"Arguments: - Lamina.DSL.config/2","doc":"- `config_key` - the name of the new configuration parameter to define.\n  - `block` - a \"do block\" which will be evaluated in the context of the\n    `Lamina.DSL.Config` module.","ref":"Lamina.DSL.html#config/2-arguments"},{"type":"macro","title":"Example - Lamina.DSL.config/2","doc":"```elixir\nconfig :listen_port do\n  cast(&Lamina.Cast.to_integer/1)\n  validate(&is_integer/1)\nend\n```","ref":"Lamina.DSL.html#config/2-example"},{"type":"macro","title":"Lamina.DSL.provider/1","doc":"Defines a provider for the configuration system.\n\nThe same as `provider/2`, however passes an empty list to the provider's\n`init/1` function.","ref":"Lamina.DSL.html#provider/1"},{"type":"macro","title":"Lamina.DSL.provider/2","doc":"Defines a provider for the configuration system.","ref":"Lamina.DSL.html#provider/2"},{"type":"macro","title":"Arguments: - Lamina.DSL.provider/2","doc":"- `module` - the name of a module which implements the `Lamina.Provider`\n  behaviour.\n- `options` - a keyword list of options to be passed to the provider's\n  `init/1` function.","ref":"Lamina.DSL.html#provider/2-arguments"},{"type":"macro","title":"Example - Lamina.DSL.provider/2","doc":"```elixir\nprovider(Lamina.Provider.Env, prefix: \"HTTP\")\n```","ref":"Lamina.DSL.html#provider/2-example"},{"type":"module","title":"Lamina.DSL.Config","doc":"Defines the macros used inside the `config` macro.","ref":"Lamina.DSL.Config.html"},{"type":"macro","title":"Lamina.DSL.Config.cast/1","doc":"Specify a transformation function to cast the value to the final required\ntype.\n\nSome configuration providers (most notably `Env`) are only able to return\nstrings, so it can be necessary to modify them before they're returned to the\nuser.","ref":"Lamina.DSL.Config.html#cast/1"},{"type":"macro","title":"Example - Lamina.DSL.Config.cast/1","doc":"```elixir\ndefmodule MyHttpServer.Config do\n  use Lamina\n\n  provider(Lamina.Provider.Env)\n\n  config :listen_port do\n    cast(&String.to_integer/1)\n  end\nend\n```","ref":"Lamina.DSL.Config.html#cast/1-example"},{"type":"macro","title":"Lamina.DSL.Config.validate/1","doc":"Specify a validation function to ensure that the value is valid.\n\nGives you an opportunity to ensure that the value about to be returned to the\nuser is correct.","ref":"Lamina.DSL.Config.html#validate/1"},{"type":"macro","title":"Example - Lamina.DSL.Config.validate/1","doc":"```elixir\ndefmodule MyFileReader.Config do\n  use Lamina\n\n  provider(Lamina.Provider.Env)\n\n  config :file_to_read do\n    validate(fn\n      \"/etc/password\" -> false\n      _ -> true\n    end)\n  end\nend\n```","ref":"Lamina.DSL.Config.html#validate/1-example"},{"type":"behaviour","title":"Lamina.Provider","doc":"The `Provider` behaviour is used to declare configuration providers.\n\nLamina's flexibility comes from the ability to provide adapters for may\nconfiguration sources and be able to query them in order to build a composite\nview of the system's configuration.  The `Lamina.Provider` behaviour allows\nyou to define these adapters.","ref":"Lamina.Provider.html"},{"type":"behaviour","title":"Example: - Lamina.Provider","doc":"As a minimum you must implement the `fetch_config/2` callback, which is used\nto look up individual configuration values in your storage:\n\n```elixir\ndefmodule MapProvider do\n  use Lamina.Provider\n\n  @impl true\n  def fetch_config(config_key, state) do\n    case Map.fetch(state, config_key) do\n      {:ok, value} -> {:ok, value, :volatile, state}\n      :error -> {:ok, state}\n    end\n  end\nend\n```","ref":"Lamina.Provider.html#module-example"},{"type":"callback","title":"Lamina.Provider.config_change/2","doc":"Config change callback. Optional.\n\nDepending on the semantics of the provider, it may be necessary for the\nprovider to proactively notify the Lamina server that a configuration value\nhas changed.  After initialisation the server will call this callback on all\nproviders with a function as it's first argument.\n\nCalling this function will cause the server to immediately call the\n`fetch_config/2` callback.","ref":"Lamina.Provider.html#c:config_change/2"},{"type":"callback","title":"Lamina.Provider.fetch_config/2","doc":"Called by Lamina when it wants to retrieve a configuration value.","ref":"Lamina.Provider.html#c:fetch_config/2"},{"type":"callback","title":"Arguments - Lamina.Provider.fetch_config/2","doc":"- `config_key` - the name of the configuration value to be returned.\n  - `state` - the current provider state.","ref":"Lamina.Provider.html#c:fetch_config/2-arguments"},{"type":"callback","title":"Return values - Lamina.Provider.fetch_config/2","doc":"There are three possible return values:\n\n  1. A successful retrieval of a configuration value, including a lifetime\n     indicator.\n  2. Inability to supply a configuration value, which is not an error.\n  3. An error.","ref":"Lamina.Provider.html#c:fetch_config/2-return-values"},{"type":"callback","title":"Lifetime indicators - Lamina.Provider.fetch_config/2","doc":"When returning a configuration value, Lamina expects a provider to provide\nadditional information about how long it can expect the value to be valid for.\n\nValid lifetime indicators are:\n\n  1. `:static` indicating that the value can never change.  This is most\n     likely for a default value, or one which may have been compiled in.\n  2. `:volatile` indicating that the value can potentially change every time\n     it is read.\n  3. A tuple containing a combination of a positive integer and a time unit.\n     Units are those specified by the `System.time_unit` typespec. For example\n     `{60, :second}` or `{3, :microsecond}`.  Useful for providers such as\n     [Vault](https://www.vaultproject.io/) which allow values to be leased for\n     a specific period of time.","ref":"Lamina.Provider.html#c:fetch_config/2-lifetime-indicators"},{"type":"callback","title":"Lamina.Provider.init/1","doc":"Called by Lamina when initialising a configuration provider.\n\nIf you're implementing a configuration provider, then this is your opportunity\nto do any setup work before any configuration is requested.\n\nWhen returning successfully, it is possible to include a `refresh_period`\noption, which indicates how often (in milliseconds) to refresh the\nconfiguration values of this provider.  This can be used by providers such as\n`Lamina.Provider.Env` which represent volatile data, but have no way of\ndetecting change other than polling.","ref":"Lamina.Provider.html#c:init/1"},{"type":"type","title":"Lamina.Provider.lifetime/0","doc":"","ref":"Lamina.Provider.html#t:lifetime/0"},{"type":"type","title":"Lamina.Provider.state/0","doc":"","ref":"Lamina.Provider.html#t:state/0"},{"type":"type","title":"Lamina.Provider.t/0","doc":"","ref":"Lamina.Provider.html#t:t/0"},{"type":"module","title":"Lamina.Provider.ApplicationEnv","doc":"Configuration provider for the OTP application environment.\n\nAllows you to use the configuration from your application enviroment\n(typically stored in `config/config.exs` et al inside your project).\n\nIn order to correctly find configuration for your application you must provide\nthe `otp_app` option, and possibly a second-level configuration key.\n\nThis is a very simple wrapper around `Application.get_all_env/1` and\n`Application.get_env/3`.  See the Elixir core documentation for more\ninformation.","ref":"Lamina.Provider.ApplicationEnv.html"},{"type":"module","title":"Examples - Lamina.Provider.ApplicationEnv","doc":"Given the following configurations:\n\n```elixir\nimport Config\nconfig :lamina,\n  jigga_watts: 1.21,\n  time_machine: [model: :delorean, top_speed: 88]\n```\n\nWe can retrieve the configurations from the environment:\n\n    iex> {:ok, state, _refresh_period} = ApplicationEnv.init(otp_app: :lamina)\n    ...> {:ok, 1.21, :volatile, _state} = ApplicationEnv.fetch_config(:jigga_watts, state)\n\n    iex> {:ok, state, _refresh_period} = ApplicationEnv.init(otp_app: :lamina, key: :time_machine)\n    ...> {:ok, :delorean, :volatile, _state} = ApplicationEnv.fetch_config(:model, state)","ref":"Lamina.Provider.ApplicationEnv.html#module-examples"},{"type":"type","title":"Lamina.Provider.ApplicationEnv.key_option/0","doc":"","ref":"Lamina.Provider.ApplicationEnv.html#t:key_option/0"},{"type":"type","title":"Lamina.Provider.ApplicationEnv.lifetime_option/0","doc":"","ref":"Lamina.Provider.ApplicationEnv.html#t:lifetime_option/0"},{"type":"type","title":"Lamina.Provider.ApplicationEnv.option/0","doc":"","ref":"Lamina.Provider.ApplicationEnv.html#t:option/0"},{"type":"type","title":"Lamina.Provider.ApplicationEnv.options/0","doc":"","ref":"Lamina.Provider.ApplicationEnv.html#t:options/0"},{"type":"type","title":"Lamina.Provider.ApplicationEnv.otp_app_option/0","doc":"","ref":"Lamina.Provider.ApplicationEnv.html#t:otp_app_option/0"},{"type":"type","title":"Lamina.Provider.ApplicationEnv.refresh_period_option/0","doc":"","ref":"Lamina.Provider.ApplicationEnv.html#t:refresh_period_option/0"},{"type":"type","title":"Lamina.Provider.ApplicationEnv.state/0","doc":"","ref":"Lamina.Provider.ApplicationEnv.html#t:state/0"},{"type":"module","title":"Lamina.Provider.Default","doc":"Configuration provider for static default configuration values.\n\nValues are passed in as a keyword list of options to init.","ref":"Lamina.Provider.Default.html"},{"type":"module","title":"Example - Lamina.Provider.Default","doc":"iex> {:ok, state} = Default.init(name: \"Marty McFly\")\n    ...> {:ok, \"Marty McFly\", :static, _state} = Default.fetch_config(:name, state)","ref":"Lamina.Provider.Default.html#module-example"},{"type":"module","title":"Lamina.Provider.Env","doc":"A configuration provider for retrieving information from the UNIX process\nenvironment.\n\nAllows you to use process enviroment variables as application configuration.\n\nThis is a very simple wrapper around `System.get_env/2`.  See the Elixir core\ndocumentation for more information.","ref":"Lamina.Provider.Env.html"},{"type":"module","title":"Name mangling - Lamina.Provider.Env","doc":"We use a process called \"name mangling\" to convert an atom configuration key\ninto the name of an environment variable.  The default mangler\n(`name_mangler/2`) converts the configuration key into UPPER_SNAKE_CASE with\nan optional prefix.\n\nShould you need to you can define your own mangler function and pass it as the\n`:mangler` option to this provider.","ref":"Lamina.Provider.Env.html#module-name-mangling"},{"type":"module","title":"Examples - Lamina.Provider.Env","doc":"Given the following environment variables:\n\n| Name              |   Value |\n| ----------------- | ------- |\n| PHOENIX_HTTP_PORT |    4000 |\n| BIND_ADDR         | 0.0.0.0 |\n\nWe can retrieve the configurations from the environment:\n\n    iex> {:ok, state, _} = Env.init(prefix: \"PHOENIX\")\n    ...> {:ok, \"4000\", :volatile, _state} = Env.fetch_config(:http_port, state)\n\n    iex> {:ok, state, _} = Env.init([])\n    ...> {:ok, \"0.0.0.0\", :volatile, _state} = Env.fetch_config(:bind_addr, state)","ref":"Lamina.Provider.Env.html#module-examples"},{"type":"function","title":"Lamina.Provider.Env.name_mangler/2","doc":"The default name mangler for environment variables.\n\nConverts the configuration key into an upper-case underscored string,\noptionally with a prefix.","ref":"Lamina.Provider.Env.html#name_mangler/2"},{"type":"function","title":"Examples: - Lamina.Provider.Env.name_mangler/2","doc":"iex> Env.name_mangler(:http_port, %{})\n    \"HTTP_PORT\"\n\n    iex> Env.name_mangler(:http_port, %{prefix: \"phoenix\"})\n    \"PHOENIX_HTTP_PORT\"","ref":"Lamina.Provider.Env.html#name_mangler/2-examples"},{"type":"type","title":"Lamina.Provider.Env.lifetime_option/0","doc":"","ref":"Lamina.Provider.Env.html#t:lifetime_option/0"},{"type":"type","title":"Lamina.Provider.Env.mangler_option/0","doc":"","ref":"Lamina.Provider.Env.html#t:mangler_option/0"},{"type":"type","title":"Lamina.Provider.Env.option/0","doc":"","ref":"Lamina.Provider.Env.html#t:option/0"},{"type":"type","title":"Lamina.Provider.Env.options/0","doc":"","ref":"Lamina.Provider.Env.html#t:options/0"},{"type":"type","title":"Lamina.Provider.Env.prefix_option/0","doc":"","ref":"Lamina.Provider.Env.html#t:prefix_option/0"},{"type":"type","title":"Lamina.Provider.Env.refresh_period_option/0","doc":"","ref":"Lamina.Provider.Env.html#t:refresh_period_option/0"},{"type":"type","title":"Lamina.Provider.Env.state/0","doc":"","ref":"Lamina.Provider.Env.html#t:state/0"},{"type":"module","title":"Lamina.Registry.PubSubRegistry","doc":"An Elixir Registry which keeps track of configuration subscribers.","ref":"Lamina.Registry.PubSubRegistry.html"},{"type":"module","title":"Lamina.Registry.ServerRegistry","doc":"An Elixir Registry which keeps track of `Lamina.Server` processes.","ref":"Lamina.Registry.ServerRegistry.html"},{"type":"function","title":"Lamina.Registry.ServerRegistry.all_servers/0","doc":"Returns a list of all the currently running configuration servers on this\nsystem.","ref":"Lamina.Registry.ServerRegistry.html#all_servers/0"},{"type":"function","title":"Lamina.Registry.ServerRegistry.lookup/1","doc":"","ref":"Lamina.Registry.ServerRegistry.html#lookup/1"},{"type":"function","title":"Lamina.Registry.ServerRegistry.register/2","doc":"","ref":"Lamina.Registry.ServerRegistry.html#register/2"},{"type":"module","title":"Lamina.Server","doc":"The Lamina configuration server.\n\nThis server is the owner of the ETS table used to store the configuration\ninformation and handles providers and their states as well as timers for\nrefreshing and expiring the configuration.","ref":"Lamina.Server.html"},{"type":"function","title":"Lamina.Server.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Lamina.Server.html#child_spec/1"},{"type":"function","title":"Lamina.Server.get/2","doc":"Retrieve a configuration value.\n\nThis is the function which is called by the dynamically generated\nconfiguration functions in the configuration module, ie calling\n`MyHttpServer.Config.listen_port()` and\n`Lamina.Server.get(MyHttpServer.Config, :listen_port)` are functionally\nidentical.\n\nLooks the server pid and ETS table up in Lamina registry, then queries the\ntable for the best match configuration value.  In most cases this should allow\nthe configuration value to be returned quickly without having to wait for a\nGenServer round-trip as both `Registry` and the server use ETS tables with\nread concurrency enabled.\n\nThe only case where a call needs to be made to the server is when the best\nconfiguration value is marked as `:volatile` by the provider - ie the value\ncan theoretically be different each time it's called.  This is the case for\nthe `Env` and `ApplicationEnv` providers, as both of these provide wrappers\naround a potentially volatile store.","ref":"Lamina.Server.html#get/2"},{"type":"function","title":"Example - Lamina.Server.get/2","doc":"iex> Server.get(MyHttpServer.Config, :listen_port)\n    ...> {:ok, 4000}","ref":"Lamina.Server.html#get/2-example"},{"type":"function","title":"Lamina.Server.get!/2","doc":"A raising version of `get/2`.","ref":"Lamina.Server.html#get!/2"},{"type":"function","title":"Lamina.Server.init/1","doc":"Initialise the configuration server.\n\nInitialises the configuration server by calling the callbacks defined by the\n`Lamina` behaviour on the config module, starts all the module's providers and\nloads all the configuration.\n\nIf any step in this process fails the server will fail to start.","ref":"Lamina.Server.html#init/1"},{"type":"type","title":"Lamina.Server.config_key/0","doc":"","ref":"Lamina.Server.html#t:config_key/0"},{"type":"type","title":"Lamina.Server.lifetime/0","doc":"","ref":"Lamina.Server.html#t:lifetime/0"},{"type":"type","title":"Lamina.Server.provider/0","doc":"","ref":"Lamina.Server.html#t:provider/0"},{"type":"type","title":"Lamina.Server.provider_state/0","doc":"","ref":"Lamina.Server.html#t:provider_state/0"},{"type":"type","title":"Lamina.Server.state/0","doc":"","ref":"Lamina.Server.html#t:state/0"},{"type":"type","title":"Lamina.Server.value/0","doc":"","ref":"Lamina.Server.html#t:value/0"},{"type":"module","title":"Lamina.Server.ConfigModule","doc":"Wrapper around access to a Lamina configuration module","ref":"Lamina.Server.ConfigModule.html"},{"type":"function","title":"Lamina.Server.ConfigModule.assert_lamina_module/1","doc":"Ensure that the module in question actually implements the Lamina behaviour.","ref":"Lamina.Server.ConfigModule.html#assert_lamina_module/1"},{"type":"function","title":"Lamina.Server.ConfigModule.cast/3","doc":"Call the Lamina callback on the configuration module to cast a configuration\nvalue.","ref":"Lamina.Server.ConfigModule.html#cast/3"},{"type":"function","title":"Lamina.Server.ConfigModule.config_change/4","doc":"Call the `config_change/3` callback on the configuration module.","ref":"Lamina.Server.ConfigModule.html#config_change/4"},{"type":"function","title":"Lamina.Server.ConfigModule.config_keys/1","doc":"Call the Lamina callback on the configuration module and return a list of\nconfig keys.","ref":"Lamina.Server.ConfigModule.html#config_keys/1"},{"type":"function","title":"Lamina.Server.ConfigModule.providers/1","doc":"Call the Lamina callback on the configuration module and return the providers.","ref":"Lamina.Server.ConfigModule.html#providers/1"},{"type":"function","title":"Lamina.Server.ConfigModule.validate/3","doc":"Call the Lamina callback on the configuration module to validate a\nconfiguration value.","ref":"Lamina.Server.ConfigModule.html#validate/3"},{"type":"module","title":"Lamina.Server.ConfigValue","doc":"A wrapper around an individual configuration value.\n\nIt makes the code much simpler. That is all.","ref":"Lamina.Server.ConfigValue.html"},{"type":"function","title":"Lamina.Server.ConfigValue.cast/1","doc":"Perform a value cast using the ConfigValue's module.","ref":"Lamina.Server.ConfigValue.html#cast/1"},{"type":"function","title":"Lamina.Server.ConfigValue.expired?/1","doc":"Has the ConfigValue expired?","ref":"Lamina.Server.ConfigValue.html#expired?/1"},{"type":"function","title":"Lamina.Server.ConfigValue.init/6","doc":"Initialise a new ConfigValue.","ref":"Lamina.Server.ConfigValue.html#init/6"},{"type":"function","title":"Lamina.Server.ConfigValue.validate/1","doc":"Perform a value validation using the ConfigValue's module.","ref":"Lamina.Server.ConfigValue.html#validate/1"},{"type":"type","title":"Lamina.Server.ConfigValue.config_key/0","doc":"","ref":"Lamina.Server.ConfigValue.html#t:config_key/0"},{"type":"type","title":"Lamina.Server.ConfigValue.lifetime/0","doc":"","ref":"Lamina.Server.ConfigValue.html#t:lifetime/0"},{"type":"type","title":"Lamina.Server.ConfigValue.provider/0","doc":"","ref":"Lamina.Server.ConfigValue.html#t:provider/0"},{"type":"type","title":"Lamina.Server.ConfigValue.provider_index/0","doc":"","ref":"Lamina.Server.ConfigValue.html#t:provider_index/0"},{"type":"type","title":"Lamina.Server.ConfigValue.t/0","doc":"","ref":"Lamina.Server.ConfigValue.html#t:t/0"},{"type":"module","title":"Lamina.Server.Impl","doc":"Separates the functional implementation from the stateful GenServer.","ref":"Lamina.Server.Impl.html"},{"type":"function","title":"Lamina.Server.Impl.get_latest_value/3","doc":"Query `provider` for the latest `config_key`.\n\nIf the provider has a new value, and it is different to the old value, then broadcast the change to subscribers, and call the config module's `config_change/3` callback.","ref":"Lamina.Server.Impl.html#get_latest_value/3"},{"type":"function","title":"Lamina.Server.Impl.init/1","doc":"Initialise the table and server state from the provided options.","ref":"Lamina.Server.Impl.html#init/1"},{"type":"function","title":"Lamina.Server.Impl.maybe_get_latest/2","doc":"If `config_value` is volatile, then ask the Lamina server to make sure that it\nhas the freshest information.","ref":"Lamina.Server.Impl.html#maybe_get_latest/2"},{"type":"function","title":"Lamina.Server.Impl.refresh_provider/2","doc":"Refreshes all configuration for a provider.","ref":"Lamina.Server.Impl.html#refresh_provider/2"},{"type":"function","title":"Lamina.Server.Impl.set_provider_config_change_callbacks/1","doc":"Call `config_change/2` on all providers with a function which they can use to\ntrigger refreshes.","ref":"Lamina.Server.Impl.html#set_provider_config_change_callbacks/1"},{"type":"type","title":"Lamina.Server.Impl.config_key/0","doc":"","ref":"Lamina.Server.Impl.html#t:config_key/0"},{"type":"type","title":"Lamina.Server.Impl.provider/0","doc":"","ref":"Lamina.Server.Impl.html#t:provider/0"},{"type":"module","title":"Lamina.Server.Provider","doc":"Helps the Lamina server deal with individual providers.","ref":"Lamina.Server.Provider.html"},{"type":"type","title":"Lamina.Server.Provider.config_key/0","doc":"","ref":"Lamina.Server.Provider.html#t:config_key/0"},{"type":"type","title":"Lamina.Server.Provider.lifetime/0","doc":"","ref":"Lamina.Server.Provider.html#t:lifetime/0"},{"type":"type","title":"Lamina.Server.Provider.provider/0","doc":"","ref":"Lamina.Server.Provider.html#t:provider/0"},{"type":"type","title":"Lamina.Server.Provider.state/0","doc":"","ref":"Lamina.Server.Provider.html#t:state/0"},{"type":"module","title":"Lamina.Server.State","doc":"Defines the state of the Lamina server process.","ref":"Lamina.Server.State.html"},{"type":"type","title":"Lamina.Server.State.provider_state/0","doc":"","ref":"Lamina.Server.State.html#t:provider_state/0"},{"type":"type","title":"Lamina.Server.State.t/0","doc":"","ref":"Lamina.Server.State.html#t:t/0"},{"type":"module","title":"Lamina.Server.Table","doc":"A wrapper around ETS for our specific use cases.","ref":"Lamina.Server.Table.html"},{"type":"function","title":"Lamina.Server.Table.expire/1","doc":"Delete any expired rows from the ETS table.","ref":"Lamina.Server.Table.html#expire/1"},{"type":"function","title":"Lamina.Server.Table.get/2","doc":"Find the most likely configuration value for a given configuration key.\n\nThis executes a gently complicated match spec against the ETS table to find a\nconfiguration value which has the highest provider priority and is not\nexpired.","ref":"Lamina.Server.Table.html#get/2"},{"type":"function","title":"Lamina.Server.Table.get/3","doc":"Get a specific configuration value from the ETS table.\n\nSearches by `config_key` and `provider` only - **does not take into account\nthe row's expiry time**.","ref":"Lamina.Server.Table.html#get/3"},{"type":"function","title":"Lamina.Server.Table.insert/2","doc":"Insert any number of configuration values into the ETS table.","ref":"Lamina.Server.Table.html#insert/2"},{"type":"function","title":"Lamina.Server.Table.new/1","doc":"Initialises a new ETS table.","ref":"Lamina.Server.Table.html#new/1"},{"type":"function","title":"Lamina.Server.Table.remove/3","doc":"Remove matching configuration values from the ETS table.","ref":"Lamina.Server.Table.html#remove/3"},{"type":"type","title":"Lamina.Server.Table.lifetime/0","doc":"","ref":"Lamina.Server.Table.html#t:lifetime/0"},{"type":"type","title":"Lamina.Server.Table.provider/0","doc":"","ref":"Lamina.Server.Table.html#t:provider/0"},{"type":"type","title":"Lamina.Server.Table.t/0","doc":"","ref":"Lamina.Server.Table.html#t:t/0"},{"type":"type","title":"Lamina.Server.Table.value/0","doc":"","ref":"Lamina.Server.Table.html#t:value/0"},{"type":"module","title":"MyHttpServer.Config","doc":"This is an example configuration module, mostly used for tests.","ref":"MyHttpServer.Config.html"},{"type":"function","title":"MyHttpServer.Config.listen_address/0","doc":"The address of the network interface upon which to bind.","ref":"MyHttpServer.Config.html#listen_address/0"},{"type":"function","title":"MyHttpServer.Config.listen_address!/0","doc":"","ref":"MyHttpServer.Config.html#listen_address!/0"},{"type":"function","title":"MyHttpServer.Config.listen_port/0","doc":"The TCP port upon which to listen for HTTP requests.","ref":"MyHttpServer.Config.html#listen_port/0"},{"type":"function","title":"MyHttpServer.Config.listen_port!/0","doc":"","ref":"MyHttpServer.Config.html#listen_port!/0"},{"type":"function","title":"MyHttpServer.Config.subscribe/1","doc":"Subscribe the calling process to notifications about configuration\nchanges.\n\nMessages in the form of `{:config_change, MyHttpServer.Config,\nconfig_key, old_value, new_value}` will be sent to the receiving process\nwhen a configuration change is detected.","ref":"MyHttpServer.Config.html#subscribe/1"},{"type":"function","title":"MyHttpServer.Config.unsubscribe/1","doc":"Unsubscribe the calling process from configuration change notifications.","ref":"MyHttpServer.Config.html#unsubscribe/1"},{"type":"exception","title":"Lamina.Error.AlreadyRegisteredError","doc":"Lamina is unable to register the configuration module.\n\nLamina uses an Elixir node-local `Registry` to keep track of all the\nconfiguration modules and their respective ETS tables.  This error indicates\nthat an attempt has been made to add the same configuration module twice.\n\nThis is most likely to happen if you have accidentally added your\nconfiguration module to a supervisor more than once.","ref":"Lamina.Error.AlreadyRegisteredError.html"},{"type":"type","title":"Lamina.Error.AlreadyRegisteredError.t/0","doc":"","ref":"Lamina.Error.AlreadyRegisteredError.html#t:t/0"},{"type":"exception","title":"Lamina.Error.ConfigNotFoundError","doc":"No configuration value was found.\n\nThis is the result of querying the ETS table for a value, but no applicable\nrecord being present.  Some possible reasons for this to happen:\n\n  - No providers returned a value for this config key.\n  - All values present in the table have expired.\n  - The config key is incorrect.","ref":"Lamina.Error.ConfigNotFoundError.html"},{"type":"type","title":"Lamina.Error.ConfigNotFoundError.t/0","doc":"","ref":"Lamina.Error.ConfigNotFoundError.html#t:t/0"},{"type":"exception","title":"Lamina.Error.InvalidValueError","doc":"A configuration value returned by a provider has failed validation.","ref":"Lamina.Error.InvalidValueError.html"},{"type":"exception","title":"Lamina.Error.NotALaminaModuleError","doc":"The module does not implement the `Lamina` behaviour.\n\nDuring startup, the Lamina server double checks that the module it's been\nasked to manage actually is a Lamina module and not just some other random\nmodule.  Really the only way you should ever see this error in the wild is if\nyou try and manually start `Lamina.Server` - which you probably shouldn't be\ndoing anyway.","ref":"Lamina.Error.NotALaminaModuleError.html"},{"type":"type","title":"Lamina.Error.NotALaminaModuleError.t/0","doc":"","ref":"Lamina.Error.NotALaminaModuleError.html#t:t/0"},{"type":"exception","title":"Lamina.Error.NotAProviderModuleError","doc":"The module does not implement the `Lamina.Provider` behaviour.\n\nDuring startup, the Lamina server double checks that the providers it's been\nasked to use actually are `Lamina.Provider` modules and not just some other\nrandom module.\n\nSome reasons you could be seeing this error:\n\n  - You have misspelt the name of the module when using the `provider` macro\n    in your configuration module (or you're missing an alias).\n  - You are developing a provider and haven't fully implemented all the\n    required callbacks.","ref":"Lamina.Error.NotAProviderModuleError.html"},{"type":"type","title":"Lamina.Error.NotAProviderModuleError.t/0","doc":"","ref":"Lamina.Error.NotAProviderModuleError.html#t:t/0"},{"type":"exception","title":"Lamina.Error.NotRegisteredError","doc":"Lamina is unable to locate the configuration module in the registry.\n\n Lamina uses an Elixir node-local `Registry` to keep track of all the\nconfiguration modules and their respective ETS tables.  This error indicates\nthat a request has been made to retrieve the pid or table of a configuration\nmodule which is not present in the registry.\n\nThis is most likely to happen if the Lamina server process has exited for some\nreason and could not be or has not yet been restarted.","ref":"Lamina.Error.NotRegisteredError.html"},{"type":"type","title":"Lamina.Error.NotRegisteredError.t/0","doc":"","ref":"Lamina.Error.NotRegisteredError.html#t:t/0"},{"type":"exception","title":"Lamina.Error.StateError","doc":"The Lamina Server is in an invalid state.\n\nThis is an internal error within `Lamina.Server` and is strongly indicitave of\na bug.  Please [open an issue][1].\n\n[1]: https://harton.dev/james/lamina/-/issues","ref":"Lamina.Error.StateError.html"},{"type":"type","title":"Lamina.Error.StateError.t/0","doc":"","ref":"Lamina.Error.StateError.html#t:t/0"},{"type":"extras","title":"README","doc":"[![Build Status](https://drone.harton.dev/api/badges/james/lamina/status.svg)](https://drone.harton.dev/james/lamina)\n[![Hex.pm](https://img.shields.io/hexpm/v/lamina.svg)](https://hex.pm/packages/lamina)\n[![Hippocratic License HL3-FULL](https://img.shields.io/static/v1?label=Hippocratic%20License&message=HL3-FULL&labelColor=5e2751&color=bc8c3d)](https://firstdonoharm.dev/version/3/0/full.html)\n\n# Lamina\n\nDynamic, runtime configuration for your Elixir app.\n\nLamina allows you to define a run-time configuration pipeline that can merge\nconfiguration from several sources. This allows the system to be reactive to\nchanges in its environment.","ref":"readme.html"},{"type":"extras","title":"Example - README","doc":"The following example defines a configuration for an imaginary HTTP server\napplication which takes it's configuration from a combination of default values,\nthe OTP application environment and system environment variables:\n\n```elixir\ndefmodule MyHttpServer.Config do\n  use Lamina\n\n  provider(Lamina.Provider.Default, listen_port: 4000, listen_address: \"0.0.0.0\")\n  provider(Lamina.Provider.ApplicationEnv, otp_app: :my_http_server, key: MyHttpServer.Endpoint)\n  provider(Lamina.Provider.Env, prefix: \"HTTP\")\n\n  config :listen_port do\n    cast(&Lamina.Cast.to_integer/1)\n\n    validate(fn\n      port when is_integer(port) and (port in [80, 443] or port >= 1000) -> true\n      _ -> false\n    end)\n  end\n\n  config :listen_address do\n    validate(fn\n      address when is_binary(address) ->\n        address\n        |> String.to_charlist()\n        |> :inet.parse_address()\n        |> case do\n          {:ok, _} -> true\n          _ -> false\n        end\n\n      _ ->\n        false\n    end)\n  end\nend\n```\n\nProvider order is preserved, such that providers added later (via the\n`provider/1` or `provider/2` macro) have more priority than their predecessors.\nThis has the effect that when more than one provider can provide a value for a\ngiven configuration item, the most preferred value will be returned.\n\nEach configuration item is defined using the `config/1` or `config/2` macro. If\nthe configuration item does not need casting to another type, nor validation\nthen just defining it with `config/1` is sufficient. In some cases it is\nnecessary to provide additional casting or validating functions. They can be\nprovided by passing a block containing the `cast/1` or `validate/1` macros.\n\nMake sure that you add your configuration module to your application's\nsupervisor tree **before** any processes that rely on it's information. Lamina\nwill fail to start or shutdown on any errors it encounters.","ref":"readme.html#example"},{"type":"extras","title":"Lifetimes - README","doc":"All configuration items in Lamina are explicitly marked with a lifetime, which\nmust be specified by the configuration provider when returning values. The\nsemantics are as follows:\n\n- `:volatile` - a configuration that could potentially be different every time\n  it is read. Volatile configuration items are returned by the `ApplicationEnv`\n  and `Env` providers.\n- `:static` - a configuration value that is not going to change until the\n  provider changes it. Static configuration items are returned by the `Default`\n  provider, but could also be used for a configuration provider that notifies\n  the system of configuration changes in some way.\n- `{non_neg_integer(), System.time_unit()}` - a value that has a specific expiry\n  time. This may be used for a configuration source that has explicit leases on\n  values (ala [Vault](https://www.vaultproject.io/) or a value for which\n  querying is expensive, and providing an expiry would effectively cache it.","ref":"readme.html#lifetimes"},{"type":"extras","title":"Querying - README","doc":"When asked to retrieve a configuration value, Lamina queries it's ETS table\nusing the following query plan; values for which there is no expiry, or which\nhave not yet expired, ordered by provider weight, descending. It only ever\nreturns a single row.\n\nIf the returned row is marked as `:volatile` then the configuration provider is\nimmediately queried for a new value, meaning that these requests will pay the\ncost of a `GenServer.call/3` to ensure freshness. If this is an issue then you\nshould consider changing the provider lifetime to use an expiry. The\n`ApplicationEnv` and `Env` providers have a configuration option to do this. If\nyou are the developer of a volatile provider, it is strongly suggested that you\nprovide for this use case.","ref":"readme.html#querying"},{"type":"extras","title":"Server configuration - README","doc":"The following options can be passed to the `use Lamina` macro, although it's\nprobably advisable to leave them as their defaults.\n\n- `gc_timeout: pos_integer()` - how long the server should be idle before\n  removing expired configuration from the ETS table in milliseconds. Defaults\n  to 3000.\n- `ttl_refresh_fraction: float` - when presented with a configuration value\n  which has an expiry, the server queues a refresh at some point prior to the\n  value expiring, in order to avoid having missing configuration. Setting this\n  to a value between `0` and `1` specifies the proportion of the expiry time to\n  wait before attempting to refresh the value. Defaults to `0.95`.","ref":"readme.html#server-configuration"},{"type":"extras","title":"Configuration subscriptions - README","doc":"Lamina defines a `subscribe/1` and `unsubscribe/1` function on each\nconfiguration module, which uses a `Registry` to handle pub-sub for\nconfiguration changes.\n\nThis allows your processes to subscribe to configuration changes and update or\nrestart any services they provide.\n\n#","ref":"readme.html#configuration-subscriptions"},{"type":"extras","title":"Example - README","doc":"For example, a simple HTTP server which changes it's listen port in response to\na configuration change:\n\n```elixir\ndefmodule MyHttpServer.Cowboy do\n  use GenServer\n  alias Plug.Cowboy\n  alias MyHttpServer.{Config, Plug}\n\n  def init(_) do\n    with {:ok, port} <- Config.listen_port(),\n        {:ok, srv} <- Cowboy.http(Plug, [], port: port),\n        :ok <- Config.subscribe(:listen_port) do\n      {:ok, %{srv: srv, port: port}}\n    end\n  end\n\n  def handle_info({:config_change, Config, :listen_port, _old_port, new_port}, %{\n        port: current_port\n      })\n      when new_port != current_port do\n    with :ok <- Cowboy.shutdown(Plug.HTTP),\n        {:ok, srv} <- Cowboy.http(Plug, [], port: new_port) do\n      {:noreply, %{port: new_port, srv: srv}}\n    else\n      {:error, reason} -> {:stop, reason, nil}\n    end\n  end\n\n  def handle_info({:config_change, _, _, _}, state), do: {:noreply, state}\nend\n```","ref":"readme.html#example"},{"type":"extras","title":"Installation - README","doc":"Lamina is [available in Hex](https://hex.pm/packages/lamina), the package can be installed\nby adding `lamina` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:lamina, \"~> 0.4.2\"}\n  ]\nend\n```\n\nDocumentation for the latest release can be found on\n[HexDocs](https://hexdocs.pm/lamina) and for the `main` branch on\n[docs.harton.nz](https://docs.harton.nz/james/lamina).","ref":"readme.html#installation"},{"type":"extras","title":"Github Mirror - README","doc":"This repository is mirrored [on Github](https://github.com/jimsynz/lamina)\nfrom it's primary location [on my Forgejo instance](https://harton.dev/james/lamina).\nFeel free to raise issues and open PRs on Github.","ref":"readme.html#github-mirror"},{"type":"extras","title":"License - README","doc":"This software is licensed under the terms of the\n[HL3-FULL](https://firstdonoharm.dev), see the `LICENSE.md` file included with\nthis package for the terms.\n\nThis license actively proscribes this software being used by and for some\nindustries, countries and activities. If your usage of this software doesn't\ncomply with the terms of this license, then [contact me](mailto:james@harton.nz)\nwith the details of your use-case to organise the purchase of a license - the\ncost of which may include a donation to a suitable charity or NGO.","ref":"readme.html#license"},{"type":"extras","title":"Change Log","doc":"# Change Log\n\nAll notable changes to this project will be documented in this file.\nSee [Conventional Commits](Https://conventionalcommits.org) for commit guidelines.\n\n<!-- changelog -->\n\n## [v0.4.2](https://harton.dev/james/lamina/compare/v0.4.1...v0.4.2) (2024-03-14)\n\n\n\n\n#","ref":"changelog.html"},{"type":"extras","title":"Bug Fixes: - Change Log","doc":"* Fix credo miswarnings.\n\n## [v0.4.1](https://harton.dev/james/lamina/compare/v0.4.0...v0.4.1) (2024-02-23)\n\n\n\n\n## [v0.4.0](https://harton.dev/james/lamina/compare/v0.3.0...v0.4.0) (2021-09-28)\n\n#","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"Features: - Change Log","doc":"- allow discovery of all running configuration servers.\n\n## [v0.3.0](https://harton.dev/james/lamina/compare/v0.2.3...v0.3.0) (2021-09-28)\n\n#","ref":"changelog.html#features"},{"type":"extras","title":"Features: - Change Log","doc":"- Add `Cast.to_boolean/1` since it's very common.\n\n## [v0.2.3](https://harton.dev/james/lamina/compare/v0.2.2...v0.2.3) (2021-07-03)\n\n#","ref":"changelog.html#features"},{"type":"extras","title":"Improvements: - Change Log","doc":"- add Cast.to_atom/1 and Cast.to_string/1.\n\n## [v0.2.2](https://harton.dev/james/lamina/compare/v0.2.1...v0.2.2) (2021-05-10)\n\n#","ref":"changelog.html#improvements"},{"type":"extras","title":"Bug Fixes: - Change Log","doc":"- dsl: fix a compiler warning about an unused module attribute.\n\n## [v0.2.1](https://harton.dev/james/lamina/compare/v0.2.0...v0.2.1) (2021-05-09)\n\n#","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"Improvements: - Change Log","doc":"- added documentation and refactored for a release.\n\n## [v0.2.0](https://harton.dev/james/lamina/compare/v0.1.0...v0.2.0) (2021-05-07)\n\n#","ref":"changelog.html#improvements"},{"type":"extras","title":"Features: - Change Log","doc":"- add configuration server and default providers.\n\n## [v0.1.0](https://harton.dev/james/lamina/compare/v0.1.0...v0.1.0) (2021-04-28)","ref":"changelog.html#features"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}